// Copyright 2022 The Sigstore Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package ca

import (
	"crypto/x509"
	"reflect"
	"testing"

	"github.com/sigstore/fulcio/pkg/test"
	"github.com/sigstore/sigstore/pkg/cryptoutils"
)

func TestSignerCerts(t *testing.T) {
	rootCert, rootKey, _ := test.GenerateRootCA()
	s := SignerCerts{Certs: []*x509.Certificate{rootCert}, Signer: rootKey}

	certs, signer := s.GetSignerWithChain()
	if err := cryptoutils.EqualKeys(signer.Public(), rootKey.Public()); err != nil {
		t.Fatalf("keys between CA and signer do not match: %v", err)
	}
	if !reflect.DeepEqual(certs, []*x509.Certificate{rootCert}) {
		t.Fatalf("expected certificate chains to match")
	}
}

func TestSignerCertsMutex(t *testing.T) {
	rootCert, rootKey, _ := test.GenerateRootCA()
	s := SignerCertsMutex{Certs: []*x509.Certificate{rootCert}, Signer: rootKey}

	certs, signer := s.GetSignerWithChain()
	if err := cryptoutils.EqualKeys(signer.Public(), rootKey.Public()); err != nil {
		t.Fatalf("keys between CA and signer do not match: %v", err)
	}
	if !reflect.DeepEqual(certs, []*x509.Certificate{rootCert}) {
		t.Fatalf("expected certificate chains to match")
	}
}
