//
// Copyright 2022 The Sigstore Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.10
// 	protoc        v6.33.1
// source: fulcio_legacy.proto

package legacy

import (
	_ "github.com/grpc-ecosystem/grpc-gateway/v2/protoc-gen-openapiv2/options"
	_ "google.golang.org/genproto/googleapis/api/annotations"
	httpbody "google.golang.org/genproto/googleapis/api/httpbody"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	emptypb "google.golang.org/protobuf/types/known/emptypb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type CreateSigningCertificateRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// The public key to be stored in the requested certificate
	//
	// Deprecated: Marked as deprecated in fulcio_legacy.proto.
	PublicKey *PublicKey `protobuf:"bytes,1,opt,name=publicKey,proto3" json:"publicKey,omitempty"`
	// Proof that the client possesses the private key
	//
	// Deprecated: Marked as deprecated in fulcio_legacy.proto.
	SignedEmailAddress []byte `protobuf:"bytes,2,opt,name=signedEmailAddress,proto3" json:"signedEmailAddress,omitempty"`
	// Optional: PKCS#10 PEM-encoded certificate signing request
	// Contains the public key to be stored in the requested
	// certificate. All other CSR fields are ignored. Since
	// the CSR is self-signed, it also acts as a proof of
	// possession of the private key.
	//
	// In particular, the CSR's subject name is not verified, or tested for
	// compatibility with its specified X.509 name type (e.g. email address).
	//
	// Deprecated: Marked as deprecated in fulcio_legacy.proto.
	CertificateSigningRequest []byte `protobuf:"bytes,3,opt,name=certificateSigningRequest,proto3" json:"certificateSigningRequest,omitempty"`
	unknownFields             protoimpl.UnknownFields
	sizeCache                 protoimpl.SizeCache
}

func (x *CreateSigningCertificateRequest) Reset() {
	*x = CreateSigningCertificateRequest{}
	mi := &file_fulcio_legacy_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CreateSigningCertificateRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CreateSigningCertificateRequest) ProtoMessage() {}

func (x *CreateSigningCertificateRequest) ProtoReflect() protoreflect.Message {
	mi := &file_fulcio_legacy_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CreateSigningCertificateRequest.ProtoReflect.Descriptor instead.
func (*CreateSigningCertificateRequest) Descriptor() ([]byte, []int) {
	return file_fulcio_legacy_proto_rawDescGZIP(), []int{0}
}

// Deprecated: Marked as deprecated in fulcio_legacy.proto.
func (x *CreateSigningCertificateRequest) GetPublicKey() *PublicKey {
	if x != nil {
		return x.PublicKey
	}
	return nil
}

// Deprecated: Marked as deprecated in fulcio_legacy.proto.
func (x *CreateSigningCertificateRequest) GetSignedEmailAddress() []byte {
	if x != nil {
		return x.SignedEmailAddress
	}
	return nil
}

// Deprecated: Marked as deprecated in fulcio_legacy.proto.
func (x *CreateSigningCertificateRequest) GetCertificateSigningRequest() []byte {
	if x != nil {
		return x.CertificateSigningRequest
	}
	return nil
}

type PublicKey struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// The cryptographic algorithm to use with the key material
	//
	// Deprecated: Marked as deprecated in fulcio_legacy.proto.
	Algorithm string `protobuf:"bytes,1,opt,name=algorithm,proto3" json:"algorithm,omitempty"`
	// PKIX, ASN.1 DER or PEM-encoded public key. PEM is typically
	// of type PUBLIC KEY.
	//
	// Deprecated: Marked as deprecated in fulcio_legacy.proto.
	Content       []byte `protobuf:"bytes,2,opt,name=content,proto3" json:"content,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PublicKey) Reset() {
	*x = PublicKey{}
	mi := &file_fulcio_legacy_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PublicKey) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PublicKey) ProtoMessage() {}

func (x *PublicKey) ProtoReflect() protoreflect.Message {
	mi := &file_fulcio_legacy_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PublicKey.ProtoReflect.Descriptor instead.
func (*PublicKey) Descriptor() ([]byte, []int) {
	return file_fulcio_legacy_proto_rawDescGZIP(), []int{1}
}

// Deprecated: Marked as deprecated in fulcio_legacy.proto.
func (x *PublicKey) GetAlgorithm() string {
	if x != nil {
		return x.Algorithm
	}
	return ""
}

// Deprecated: Marked as deprecated in fulcio_legacy.proto.
func (x *PublicKey) GetContent() []byte {
	if x != nil {
		return x.Content
	}
	return nil
}

var File_fulcio_legacy_proto protoreflect.FileDescriptor

const file_fulcio_legacy_proto_rawDesc = "" +
	"\n" +
	"\x13fulcio_legacy.proto\x12\x1adev.sigstore.fulcio.v1beta\x1a\x1cgoogle/api/annotations.proto\x1a\x1fgoogle/api/field_behavior.proto\x1a\x19google/api/httpbody.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a.protoc-gen-openapiv2/options/annotations.proto\"\xec\x01\n" +
	"\x1fCreateSigningCertificateRequest\x12K\n" +
	"\tpublicKey\x18\x01 \x01(\v2%.dev.sigstore.fulcio.v1beta.PublicKeyB\x06\xe2A\x01\x01\x18\x01R\tpublicKey\x126\n" +
	"\x12signedEmailAddress\x18\x02 \x01(\fB\x06\xe2A\x01\x01\x18\x01R\x12signedEmailAddress\x12D\n" +
	"\x19certificateSigningRequest\x18\x03 \x01(\fB\x06\xe2A\x01\x01\x18\x01R\x19certificateSigningRequest\"O\n" +
	"\tPublicKey\x12 \n" +
	"\talgorithm\x18\x01 \x01(\tB\x02\x18\x01R\talgorithm\x12 \n" +
	"\acontent\x18\x02 \x01(\fB\x06\xe2A\x01\x02\x18\x01R\acontent2\xf8\x01\n" +
	"\x02CA\x12\x90\x01\n" +
	"\x18CreateSigningCertificate\x12;.dev.sigstore.fulcio.v1beta.CreateSigningCertificateRequest\x1a\x14.google.api.HttpBody\"!\x82\xd3\xe4\x93\x02\x18:\x01*\"\x13/api/v1/signingCert\x88\x02\x01\x12_\n" +
	"\x12GetRootCertificate\x12\x16.google.protobuf.Empty\x1a\x14.google.api.HttpBody\"\x1b\x82\xd3\xe4\x93\x02\x12\x12\x10/api/v1/rootCert\x88\x02\x01B\xa1\x03\x92A\xb8\x02\x12\xc0\x01\n" +
	"\rFulcio Legacy\"\\\n" +
	"\x17sigstore Fulcio project\x12\"https://github.com/sigstore/fulcio\x1a\x1dsigstore-dev@googlegroups.com*J\n" +
	"\x12Apache License 2.0\x124https://github.com/sigstore/fulcio/blob/main/LICENSE2\x051.0.0\x1a\x13fulcio.sigstore.dev*\x01\x012\x10application/json:\x10application/jsonr7\n" +
	"\x11More about Fulcio\x12\"https://github.com/sigstore/fulcio\n" +
	"\x1adev.sigstore.fulcio.v1betaB\vFulcioProtoP\x01Z8github.com/sigstore/fulcio/pkg/generated/protobuf/legacyb\x06proto3"

var (
	file_fulcio_legacy_proto_rawDescOnce sync.Once
	file_fulcio_legacy_proto_rawDescData []byte
)

func file_fulcio_legacy_proto_rawDescGZIP() []byte {
	file_fulcio_legacy_proto_rawDescOnce.Do(func() {
		file_fulcio_legacy_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_fulcio_legacy_proto_rawDesc), len(file_fulcio_legacy_proto_rawDesc)))
	})
	return file_fulcio_legacy_proto_rawDescData
}

var file_fulcio_legacy_proto_msgTypes = make([]protoimpl.MessageInfo, 2)
var file_fulcio_legacy_proto_goTypes = []any{
	(*CreateSigningCertificateRequest)(nil), // 0: dev.sigstore.fulcio.v1beta.CreateSigningCertificateRequest
	(*PublicKey)(nil),                       // 1: dev.sigstore.fulcio.v1beta.PublicKey
	(*emptypb.Empty)(nil),                   // 2: google.protobuf.Empty
	(*httpbody.HttpBody)(nil),               // 3: google.api.HttpBody
}
var file_fulcio_legacy_proto_depIdxs = []int32{
	1, // 0: dev.sigstore.fulcio.v1beta.CreateSigningCertificateRequest.publicKey:type_name -> dev.sigstore.fulcio.v1beta.PublicKey
	0, // 1: dev.sigstore.fulcio.v1beta.CA.CreateSigningCertificate:input_type -> dev.sigstore.fulcio.v1beta.CreateSigningCertificateRequest
	2, // 2: dev.sigstore.fulcio.v1beta.CA.GetRootCertificate:input_type -> google.protobuf.Empty
	3, // 3: dev.sigstore.fulcio.v1beta.CA.CreateSigningCertificate:output_type -> google.api.HttpBody
	3, // 4: dev.sigstore.fulcio.v1beta.CA.GetRootCertificate:output_type -> google.api.HttpBody
	3, // [3:5] is the sub-list for method output_type
	1, // [1:3] is the sub-list for method input_type
	1, // [1:1] is the sub-list for extension type_name
	1, // [1:1] is the sub-list for extension extendee
	0, // [0:1] is the sub-list for field type_name
}

func init() { file_fulcio_legacy_proto_init() }
func file_fulcio_legacy_proto_init() {
	if File_fulcio_legacy_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_fulcio_legacy_proto_rawDesc), len(file_fulcio_legacy_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   2,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_fulcio_legacy_proto_goTypes,
		DependencyIndexes: file_fulcio_legacy_proto_depIdxs,
		MessageInfos:      file_fulcio_legacy_proto_msgTypes,
	}.Build()
	File_fulcio_legacy_proto = out.File
	file_fulcio_legacy_proto_goTypes = nil
	file_fulcio_legacy_proto_depIdxs = nil
}
