// Copyright (c) 2015, Gengo, Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//
//     * Neither the name of Gengo, Inc. nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

syntax = "proto3";

package grpc.gateway.protoc_gen_openapiv2.options;

import "google/protobuf/descriptor.proto";
import "protoc-gen-openapiv2/options/openapiv2.proto";

option go_package = "github.com/grpc-ecosystem/grpc-gateway/v2/protoc-gen-openapiv2/options";

extend google.protobuf.FileOptions {
  // ID assigned by protobuf-global-extension-registry@google.com for gRPC-Gateway project.
  //
  // All IDs are the same, as assigned. It is okay that they are the same, as they extend
  // different descriptor messages.
  Swagger openapiv2_swagger = 1042;
}
extend google.protobuf.MethodOptions {
  // ID assigned by protobuf-global-extension-registry@google.com for gRPC-Gateway project.
  //
  // All IDs are the same, as assigned. It is okay that they are the same, as they extend
  // different descriptor messages.
  Operation openapiv2_operation = 1042;
}
extend google.protobuf.MessageOptions {
  // ID assigned by protobuf-global-extension-registry@google.com for gRPC-Gateway project.
  //
  // All IDs are the same, as assigned. It is okay that they are the same, as they extend
  // different descriptor messages.
  Schema openapiv2_schema = 1042;
}
extend google.protobuf.ServiceOptions {
  // ID assigned by protobuf-global-extension-registry@google.com for gRPC-Gateway project.
  //
  // All IDs are the same, as assigned. It is okay that they are the same, as they extend
  // different descriptor messages.
  Tag openapiv2_tag = 1042;
}
extend google.protobuf.FieldOptions {
  // ID assigned by protobuf-global-extension-registry@google.com for gRPC-Gateway project.
  //
  // All IDs are the same, as assigned. It is okay that they are the same, as they extend
  // different descriptor messages.
  JSONSchema openapiv2_field = 1042;
}
